/**********************************************************************/
/* echant_strat                                                       */
/*                                                                    */
/* Drawing a stratified balanced sample                               */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* echant_strat Drawing a stratified balanced sample                  */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/***********************************************************************/
/* The echant_strat macro enables :                                    */
/*    -> to perform a flight phase independantly stratum by stratum    */
/*    -> to perform a last flight phase on all the remaining units,    */
/*       and then to perform a landing phase                           */
/* We thus obtain an exact balancing on the whole population, and an   */
/* approximate balancing by stratum                                    */
/* Moreover, a sample of fixed size is drawn in each stratum : the     */
/* parameter &contr must contain the variable giving the inclusion     */
/* probabilities                                                       */
/*                                                                     */
/*    base : library containing the data base                          */
/*    data : names of the SAS tables containing the Input Data. The    */
/*           names must be spaced with blanks. Each table contains the */
/*           units of one stratum                                      */
/*    id : variable identifying the units of the population            */
/*    pi : variable of the inclusion probabilities                     */
/*    contr : names of the balancing variables, spaced with blanks     */
/*    sort : name of the SAS output table, containing the same         */
/*           variables as data, and a variable named ech equalling 1 if*/
/*           the unit belongs to the sample and 0 otherwise            */
/*                                                                     */
/* The echant_strat macro uses SAS/MACRO and SAS/IML modules           */
/*                                                                     */
/* The echant_strat macro uses the sub-macros : vol and modules_IML    */
/* (See below)                                                         */ 
/***********************************************************************/;

%macro echant_strat(base=,data=,id=,pi=,contr=,sort=);

proc iml;

%modules_IML;

/* We begin with independant flight phases in each stratum, i.e. in    */
/* each table of the parameter &data                                   */
	
%let cal=0;
%let compt=0;
%do %while(%scan(&data,&cal+1) ne ); 

	%let cal=%eval(&cal+1);
	%let tab=%scan(&data,&cal);
	
	use &base..&tab where (&pi>0 & &pi<1);
	read all var {&id} into _id;
	read all var {&contr} into _contr_;
	read all var {&pi} into _pi;
	_pistar=J(nrow(_pi),1,0.99);

	%vol(_pi=_pi,_contr_=_contr_,_pistar=_pistar);

	/* If we know the result of the sampling for every unit of the     */
	/* stratum, we just memorize this result                           */

	call symput("piv",char(ncol(loc(_pistar>1e-8 & _pistar<1-1e-8))>0));
	%let compt=%eval(&compt+&piv);

	if ncol(loc(_pistar>1e-8 & _pistar<1-1e-8))=0 then do;
		_pistar=round(_pistar);
		create &tab._sort var {_id,_pistar};append;close &tab._sort;	
	end;

	/* Otherwise, we memorize the result of the sampling for the units */
	/* concerned. The remaining units are put with other remaining     */
	/* units of the previous strata                                    */ 

	else do;

		_r=loc(_pistar>1e-8 & _pistar<1-1e-8);
		_sauv_=t(remove(_id,_r))||t(char(round(remove(_pistar,_r))));
		varnames={_id}//{_pistar};
		create &tab._sort from _sauv_ [colname=varnames];append from _sauv_;close &tab._sort;

		a&compt=_pi[_r];
		_idglo=_idglo//_id[_r];
		_piglo=_piglo//_pi[_r];	
		_pistarglo=_pistarglo//_pistar[_r];
		_contrglo_=_contrglo_//_contr_[_r,];	

	end;

	free _id _contr_ _pi;

%end;

/* If we still don't know the outcome of the sampling for some*/
/* units, we deal with them                                   */

%if (&compt>0) %then %do;

	/* We had balancing variables to ensure a fixed size in   */
	/* each stratum                                           */
 
	%if (&compt>1) %then %do;

		_prov_=J(nrow(a1),%eval(&compt-1),0);
		_piv_=insert(_prov_,a1,0,1);

		%do spi=2 %to &compt;
			_prov_=J(nrow(a&spi),%eval(&compt-1),0);
			_prov_=insert(_prov_,a&spi,0,&spi);
			_piv_=_piv_//_prov_;
		%end;

		_contrglo_=_piv_||_contrglo_[,2:ncol(_contrglo_)];
		_contrglo_=(_contrglo_#(_pistarglo/_piglo));
		
	%end;

	/* We perform a last flight phase on the remaining units  */

	%vol(_pi=_pistarglo,_contr_=_contrglo_,_pistar=_pistarglo);

	/* If we still don't know the outcome of the sampling for */
	/* some units, we deal with them                          */

	call symput("krum",char(ncol(loc(_pistarglo>1e-8 & _pistarglo<1-1e-8))));
 
	%if &krum>0 %then %do;

		/* We begin by solving, for each stratum, the         */
		/* possible issues of non integer number of units     */
		/* remaining to select                                */ 
	
		deb=1;fin=nrow(a1);
		%do der=1 %to &compt;

			som_pi=round(_pistarglo[deb:fin][+],1e-8);
			alpha=som_pi-floor(som_pi);
			if abs(alpha)>1e-8 then do;
				
				_phi=_pistarglo[deb:fin];
				_phi1=_phi;_phi2=_phi;
				yop=1;suma=0;
				do while (abs(suma-alpha)>1e-8);
					a=min(_phi[yop],alpha-suma,alpha*(1-_phi[yop])/(1-alpha));
					_phi1[yop]=_phi[yop]-a;
					_phi2[yop]=_phi[yop]+(1/alpha-1)*a;
					suma=suma+a;
					yop=yop+1;
				end;

				_phi=_phi1+(uniform(0)<alpha)*(_phi2-_phi1);
				_pistarglo[deb:fin]=_phi;

			end;
			
			deb=fin+1;fin=fin+nrow(a&der);

		%end;

		/* If we still don't know the outcome of the sampling for some*/
		/* units, we deal with them                                   */

		call symput("vic",char(ncol(loc(_pistarglo>1e-8 & _pistarglo<1-1e-8))));

		%if &vic>0 %then %do;	

			deb=1;fin=nrow(a1);
			%let compt2=0;
			%do der=1 %to &compt;

				nr=ncol(loc(_pistarglo[deb:fin]>1e-8 & _pistarglo[deb:fin]<1-1e-8));

				/* If units remain in the present stratum, we list    */
				/* all possible samples                               */

				call symput("shrek",char(nr>0));
				%let compt2=%eval(&compt2+&shrek);

				if nr>0 then do;

					piv_=loc(_pistarglo[deb:fin]>1e-8 & _pistarglo[deb:fin]<1-1e-8);
					natter=round(_pistarglo[deb:fin][piv_][+]);

					_s=J(natter,1,1)//J(nr-natter,1,0);
					_supp&compt2._=_s;	
					do while (_s[nr-natter+1:nr][+] < natter);
						j=loc(_s=0)[><];
						_s[1:j]=1-_s[1:j];
						if _s[+]=natter then do;
							_supp&compt2._=_supp&compt2._||_s;	
						end;
					end;

				end;

				deb=fin+1;fin=fin+nrow(a&der);

			%end;

			_support_=_supp1_;	
			%let man=1;
			%if (&compt2>1) %then %do %while (&man<&compt2);
				a=_support_@J(1,ncol(_supp%eval(&man+1)_),1);
				b=repeat(_supp%eval(&man+1)_,1,ncol(_support_));
				_support_=a//b;
				%let man=%eval(&man+1);
			%end;

			_support_=J(1,ncol(_support_),0)//_support_;

			/* Definition of the coefficients of the issue        */
			/* (calculation of the costs)                         */

			total_=_contrglo_[+,];
			piv=loc(_pistarglo>1e-8 & _pistarglo<1-1e-8);
	
			do i=1 to ncol(_support_);
				difftot_=(_contrglo_[piv,]#((_pistarglo[piv]-_support_[2:nrow(_support_),i])/_piglo[piv]))[+,];
				_support_[1,i]=(difftot_/total_)[##];
			end;

			/* Building of the table used for the linear           */
			/* optimisation (use of the LP proc)                   */

			nr=nrow(_support_)-1;
			nech=ncol(_support_);
			piv1=repeat("cons",nr+1,1);piv2=left(char(t(1:nr+1)));
			_row_={"obj"}//concat(piv1,piv2)//{"lowerb"};
			varnames={_row_};
			create _row_ from _row_ [colname=varnames];append from _row_;close _row_;

			_support_=_support_//J(1,nech,1)//J(1,nech,0);
			varnames=concat(repeat("x",1,nech),left(char(1:nech)));
			create _cout_ from _support_ [colname=varnames];append from _support_;close _cout_;

			_type_={"min"}//repeat("eq",nr+1,1)//{"lowerbd"};
			varnames={_type_};
			create _type_ from _type_ [colname=varnames];append from _type_;close _type_;

			_rhs_={.}//_pistarglo[piv]//{1,.};
			varnames={_rhs_};
			create _rhs_ from _rhs_ [colname=varnames];append from _rhs_;close _rhs_;

			create _pistarglo from _pistarglo;append from _pistarglo;close _pistarglo;
			create _idglo from _idglo;append from _idglo;close _idglo;

			quit;

			/* Solution of the optimisation problem               */ 
			data lptable;merge _row_ _cout_ _type_ _rhs_;run;
			proc lp data=lptable primalout=suppmin;run;
			/* Back into SAS IML                                  */ 
			proc iml;

			use suppmin;read all var {_value_} where (_TYPE_="LOWERBD") into _patter;
			use _pistarglo;read all into _pistarglo;
			use _idglo;read all var _char_ into _idglo;
			use _cout_;read all into _cout_;

			/* Selection of a sample according to the optimal     */
			/* sampling design                                    */

			_support_=loc(abs(_patter)>1e-8);
			_support_=t(_support_)||cusum(_patter[_support_]);
			alea=uniform(0);
			s=_support_[loc(_support_[,2]>alea)[1],1];
			_s_=_cout_[2:nrow(_cout_)-2,s];

			_sauv_=_idglo||char(round(_pistarglo));

			varnames={_id}//{_pistar};
			create reliquat from _sauv_ [colname=varnames];append from _sauv_;close reliquat;

		%end;

	%end;

	/* End of the landing phase                               */

%end;

/* End of the sampling                                        */

quit;

data &base..&sort(drop=_pistar);set
%let i=0;%do %while (%scan(&data,&i+1) ne );
	%let i=%eval(&i+1);%let tab=%scan(&data,&i);
	&tab._sort
%end;
reliquat;
ech=input(compress(_pistar),1.);
rename _id=&id;
run;

%mend echant_strat;

/**********************************************************************/
/* vol                                                                */
/*                                                                    */
/* Performing a flight phase                                          */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* vol Performing a flight phase                                      */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/**********************************************************************/
/* The vol macro performs a flight phase                              */
/* Input :                                                            */
/*   -> pi : vector of inclusion probabilities                        */
/*   -> contr : matrix of balancing variables                         */
/* Output :                                                           */
/*   -> pistar : vector of inclusion probabilities after the flight   */
/*               phase                                                */
/*                                                                    */
/* The vol macro uses SAS/MACRO and SAS/IML modules                   */
/*                                                                    */
/* The echant_strat macro uses the sub-macro : "modules-IML" (See     */
/* below)                                                             */ 
/**********************************************************************/;

%macro vol(_pi=,_contr_=,_pistar=);

/* Initialization of the parameters                                   */

N=nrow(&_contr_);
p=ncol(&_contr_);
k=1;
_phi=&_pi;
&_pistar=&_pi;

/* Fast flight phase                             */

if p+1<=N then do;

	_phi=&_pi[1:(p+1)];
	_r=t(1:p+1);
	_B_=(_phi##(-1))#&_contr_[1:p+1,];
	k=p+2;

	/* A first iteration                                              */

	run choix_vecteur(_B_,_phi);
	drop_=loc(_phi<1e-8 | _phi>1-1e-8);
	r=ncol(drop_);
	
	/* As long as units for which the choice has been made can        */
	/* be replaced, we go on                                          */

	do while(r<=N-k+1);

		&_pistar[_r[drop_]]=_phi[drop_];
		_r[drop_]=t(k:k+r-1);
		_phi[drop_]=&_pi[k:k+r-1];

		_B_[drop_,]=((_phi[drop_])##(-1))#&_contr_[_r[drop_],];
		k=k+r;

		run choix_vecteur(_B_,_phi);

		drop_=loc(_phi<1e-8 | _phi>1-1e-8);
		r=ncol(drop_);

	end;

	keep_=loc(_phi>1e-8 & 1-_phi>1e-8);
	&_pistar[_r[drop_]]=_phi[drop_];

end;

/* Complement of the flight phase                                     */

if (ncol(keep_)>0)|(k<=N) then do;

	/* We join the remaining units together                           */

	if ncol(keep_)=0 then do;
		keep_={1};
		_r=t(k:N);
		_phi=&_pi[k:N];
		_B_=((&_pi[k:N])##(-1))#&_contr_[k:N,]; 
	end;
	else do;
		if k>N then do;
			_r=_r[keep_];	
			_phi=_phi[keep_];
			_B_=_B_[keep_,];
		end;
		else do;
			_r=_r[keep_]//t(k:N);	
			_phi=_phi[keep_]//&_pi[k:N];
			_B_=_B_[keep_,]//(((&_pi[k:N])##(-1))#&_contr_[k:N,]); 
		end;
	end;

	rank=rang(_B_);
	
	do while ((rank<nrow(_B_))&(ncol(keep_)>0));

		run choix_vecteur(_B_,_phi);

		drop_=loc(_phi<=1e-8 | 1-_phi<=1e-8);
		keep_=loc(_phi>1e-8 & 1-_phi>1e-8);
		&_pistar[_r[drop_]]=_phi[drop_];
		if ncol(keep_)>0 then do;
				_r=_r[keep_];
				_phi=_phi[keep_];
				_B_=_B_[keep_,];
			end;
	
	end;

end;

if ncol(keep_)>0 then &_pistar[_r]=_phi;

%mend vol;

%macro modules_IML;

	/******************************************************************/
	/* Basic loop of the flight phase                                 */
	/******************************************************************/

	start choix_vecteur(_mat_,_pistar_);
		_u_=(homogen(t(_mat_)))[,1];
		_nonnul_=loc(abs(_u_)>1e-8);
		piv=_u_[_nonnul_]; 
		_lem_=(piv##(-1))#((piv>1e-8)-_pistar_[_nonnul_]);
		l1=min(_lem_);
		_lem_=(piv##(-1))#(_pistar_[_nonnul_]-(piv<1e-8));
		l2=min(_lem_);
		if (uniform(0)<(l2/(l1+l2))) then _pistar_=_pistar_+_u_*l1;
		else _pistar_=_pistar_-_u_*l2;
	finish choix_vecteur;

	/******************************************************************/
	/* Calculation of the rank of a matrix                            */
	/******************************************************************/
	start rang(_mat_);
		X=homogen(block(_mat_,0));
		return (nrow(X)-ncol(X));
	finish rang;

%mend modules_IML;














